/*******************************************

 MAIN.CPP - The KINT Interpreter
	Copyright 2002 Everett Kaser
	All rights reserved.

	Do not copy or redistribute without the
	author's explicit written permission.
 *******************************************/

/*********************  Header Files  *********************/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <mmsystem.h>

#include "resource.h"
#include "main.h"
#include "pmach.h"
#include "sysfuncs.h"

//********************************************************************

void InitScreen();
void MakePalette();
void DestroyPalette();

//********************************************************************

extern PDIBS	kBM;
extern HDC		ddc, sdc;
extern HBITMAP	oldBMs, oldBMd;
extern UINT MCIwDeviceID;

extern HPALETTE	kPal;

//********************************************************************

#define	PATHLEN	512
#define	UM_GETPATH	(WM_USER+0)

/*******************  Global Variables ********************/

HANDLE	ghInstance;

HWND	hWndMain=NULL;

WORD	VW, VH;
WORD	MinFlag=FALSE, MusicPlaying=FALSE;

BOOL	ProgramReady=FALSE;
UINT	TimerID=NULL;

char	PgmPath[PATHLEN], TmpPath[PATHLEN];
char	szAppName[] = "KINT";
char	KEname[128];

int		ScreenWidth, ScreenHeight;
WORD	ScreenWidthIndex;

int		CurCur=0, mainShowCmd;

//********************************************************************
// callback for enumerate windows to bring previous instance of this program to the foreground
// for each top-level window, get it's class name.  If it matches "our" class name, then
// send it a message to get a ptr to it's program path.  If it matches THIS program path,
// we've found the previous instance, and we bring it to the foreground.
BOOL CALLBACK FindPrevKintWnd(HWND hWnd, LPARAM lParam) {

	GetClassName(hWnd, TmpPath, PATHLEN);
	if( !strcmp(TmpPath, szAppName) ) {
		if( !strcmp((char *)SendMessage(hWnd, UM_GETPATH, NULL, NULL), PgmPath) ) {
			SetForegroundWindow(hWnd);
			return FALSE;
		}
	}
	return TRUE;
}

//********************************************************************
// Main entry point for Windows programs
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow ) {

	WNDCLASS		wc;
	MSG				msg;
	int			i;
	HDC			hDC;
	DWORD		IdleCount;

// get path program was run from and save in PgmPath
// ALL other file access by this program are relative to PgmPath.
	i = GetModuleFileName(ghInstance, PgmPath, 511);
	while( i && PgmPath[i]!='\\' && PgmPath[i]!=':' ) --i;
	if( PgmPath[i]=='\\' || PgmPath[i]==':' ) ++i;
	if( !i || PgmPath[i-1]!='\\' ) PgmPath[i++] = '\\';
	PgmPath[i] = 0;

// see if a copy of this program is already running
// if so, just activate it and kill this copy,
// else let this copy proceed.

	// convert program folder path backslashes to underscores
	for(i=0; PgmPath[i]; i++) {
		if( PgmPath[i]=='\\' ) TmpPath[i] = '_';
		else TmpPath[i] = PgmPath[i];
	}
	TmpPath[i] = 0;
	// create a Windows mutex.  if it succeeds, this is the first instance of the program
	HANDLE	hmutex = CreateMutex(NULL, FALSE, TmpPath);
	// if it fails, there's another instance already running
	if( GetLastError()==ERROR_ALREADY_EXISTS ) {
		// find the other instance and bring it to the foreground, then terminate this instance
		EnumWindows((WNDENUMPROC)FindPrevKintWnd, NULL);
		exit(0);
	}

// if a .KE filename was specified when the program was run, use it.
// Otherwise, default to MAIN.KE
	if( lpszCmdLine[0] ) strcpy(KEname, lpszCmdLine);
	else strcpy(KEname, "main.ke");

// try to load the .KE file
	if( !LoadKEFile(KEname) ) exit(0);

// register our window class
	wc.lpszClassName = szAppName;
	wc.lpfnWndProc = MainWndProc;
	wc.style = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON));
	wc.hCursor = NULL;//LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wc.lpszMenuName = NULL;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;

	RegisterClass( &wc );
	ghInstance = hInstance;

// create a color palette in case we're running in a 256-color mode
	MakePalette();

// create the two "working" screen DCs (source DC and destination DC)
// these are used heavily in SYSFUNCS.CPP module for doing drawing to kBM and elsewhere
	hDC = GetDC(NULL);
	if( NULL==(sdc=CreateCompatibleDC(hDC)) ) {
		MessageBox(NULL, "Failed creating source DC", "Oops", MB_OK);
		ReleaseDC(NULL, hDC);
		exit(0);
	}
	if( NULL==(ddc=CreateCompatibleDC(hDC)) ) {
		MessageBox(NULL, "Failed creating destination DC", "Oops", MB_OK);
		ReleaseDC(NULL, hDC);
		exit(0);
	}
	ReleaseDC(NULL, hDC);

// create the default font and the kBM (screen) bitmap.
// the KASM code does ALL of it's drawing to kBM, a memory bitmap.
// when Windows needs to update the REAL screen, the WM_PAINT message
// in this program merely BitBlt's from kBM to the screen DC.
	InitScreen();

// call the BEGIN_STARTUP code in the KE module (tell it we're starting up and
// about to create the main program window)
/****/	KEevent(KHDR_BEGIN_STARTUP, 0, 0, 0);

// create main window
	hWndMain = CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW,
		0, 0, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInstance, NULL);

// call the END_STARTUP code in the KE module (tell it to finish setting up, and
// that the main program window has now been created.
/****/	KEevent(KHDR_END_STARTUP, 0, 0, 0);

// possibly cause the window to be shown on the display.
// 'mainShowCmd' gets set by the .KE code during KHDR_END_STARTUP using
// SysGetWindowPos and SysSetWindowPos calls.
	ShowWindow(hWndMain, mainShowCmd);

// send one IDLE message to the KASM code in case it needs to do anything once
// right up front after everything else has been initialized.
/****/	KEevent(KHDR_IDLE, 0, 0, 0);

	ProgramReady = TRUE;
// this is the main "message pump" loop.  While no messages are waiting for the
// program and the IDLE call to the KASM code returns non-0, it keeps sending
// IDLE messages to the KASM code.  If the KASM code returns 0, then no more
// IDLE messages are sent until some Windows message arrives for the program.
	for(;;) {
		IdleCount = 0;
		// check to see if we can do idle work
/****/	while( !PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) && KEevent(KHDR_IDLE, IdleCount++, 0, 0) ) ;

		// either we have a message, or OnIdle returned false
		if( GetMessage(&msg, NULL, 0, 0)==TRUE ) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else break;
	}
	return msg.wParam;
}

//********************************************************************************
// This is the main MS Windows message ("event") handling function.
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {

	HBITMAP			holdbm;
	HDC				hdc, bdc;
	PAINTSTRUCT		ps;

	switch( msg ) {
	 case 0x0319:	// BACK & FWD buttons
		if( HIWORD(lParam)==0x8001 ) {	// BACK
/****/		(LRESULT)(!KEevent(KHDR_KEY_DOWN, (DWORD)0x8001, (DWORD)(lParam&0x0ffff), 0));
			return TRUE;
		} else if( HIWORD(lParam)==0x8002 ) {	// FWD
/****/		(LRESULT)(!KEevent(KHDR_KEY_DOWN, (DWORD)0x8002, (DWORD)(lParam&0x0ffff), 0));
			return TRUE;
		}
		break;
	 case UM_GETPATH:
	// our own custom message to get the path to the other instances of
	// the program, so that we can figure out if programs by the same
	// name are actually the same program or not.  (ie, this same interpreter
	// might be used for several different games, in different folders.  While
	// the name of the EXE *MIGHT* be the same, the folders will be different,
	// so we check BOTH the program and AND the folder from which it was run.)
	// See the FindPrevKintWnd() function above.
		return (LRESULT)PgmPath;
	 case WM_ERASEBKGND:
	// never erase background, because we always draw everything
		return TRUE;
	 case MM_MCINOTIFY:
	// various MIDI messages, we only care about termination
		switch( wParam ) {
		 case MCI_NOTIFY_ABORTED:
		 case MCI_NOTIFY_FAILURE:
		 case MCI_NOTIFY_SUPERSEDED:
	 		break;
		 case MCI_NOTIFY_SUCCESSFUL:
			mciSendCommand(MCIwDeviceID, MCI_CLOSE, 0, NULL);
			MusicPlaying = 0;
/****/		KEevent(KHDR_MUSIC_DONE, 0, 0, 0);
			break;
		 default:
			break;
		}
		return 0;
	 case WM_SIZE:
	// the program window has been resized, minimized, or maximized
		if( wParam==SIZE_MINIMIZED ) {
			MinFlag = TRUE;
/****/		KEevent(KHDR_WINDOW_HIDDEN, 1, 0, 0);
		} else {
			RECT	rc;

/****/		if( MinFlag==TRUE ) KEevent(KHDR_WINDOW_HIDDEN, 0, 0, 0);
			GetClientRect(hWnd, &rc);
			VW = (WORD)(rc.right-rc.left);
			VH = (WORD)(rc.bottom-rc.top);
/****/		KEevent(KHDR_WINDOW_SIZE, LOWORD(lParam), HIWORD(lParam), 0);
			MinFlag = FALSE;
		}
		return 0;
	 case WM_DISPLAYCHANGE:
	// The resolution of the display has been changed.  Recreate
	// the "display" bitmap kBM, the defFont, and redraw the window.
		InitScreen();
		return 0;
	 case WM_MOVE:
	// the program window has been moved
/****/	KEevent(KHDR_WINDOW_MOVE, LOWORD(lParam), HIWORD(lParam), 0);
		return 0;
	 case WM_TIMER:
	// the timer time period has expired
/****/	if( ProgramReady ) KEevent(KHDR_TIMER, 0, 0, 0);
		return 0;
	 case WM_MENUCHAR:
		return (MNC_CLOSE<<16);	// to avoid BEEP when holding ALT down and pressing the "hotkey"
	 case WM_SYSKEYDOWN:
		/* NOTE: "menu" (alt-xxx) keys set ARG3 to 1, all other keys (except mouse buttons) set it to 0 */
		if( (wParam>=0x30 && wParam<=0x39) || (wParam>=0x41 && wParam<=0x5A) ) {
// Windows doesn't send a WM_CHAR for ALT-alpha keys, which screws up the logic!
// So, we send it manually in those cases, to keep things consistent.
/****/		(LRESULT)(!KEevent(KHDR_CHAR, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		} else {
/****/		(LRESULT)(!KEevent(KHDR_KEY_DOWN, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 1));
		}
		return 0;
	 case WM_SYSKEYUP:
		/* NOTE: "menu" (alt-xxx) keys set ARG3 to 1, all other keys (except mouse buttons) set it to 0 */
		if( (wParam>=0x30 && wParam<=0x39) || (wParam>=0x41 && wParam<=0x5A) ) {
			(LRESULT)(!KEevent(KHDR_ALPHAKEY_UP, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		} else {
/****/		(LRESULT)(!KEevent(KHDR_KEY_UP, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 1));
		}
		return 0;
	 case WM_KEYDOWN:
	// a keyboard key has been pressed
		if( (wParam>=0x30 && wParam<=0x39) || (wParam>=0x41 && wParam<=0x5A) ) {
			(LRESULT)(!KEevent(KHDR_ALPHAKEY_DOWN, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		} else {
/****/		(LRESULT)(!KEevent(KHDR_KEY_DOWN, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		}
		return 0;
	 case WM_KEYUP:
	// a keyboard key has been released
	 // KHDR_KEY_UP and KHDR_KEY_DOWN *MUST* be sent for at least SHIFT, CTL, ALT,
	 // TAB, and ENTER, or menus and dialogs will be broken (so far as the keyboard goes).
		if( (wParam>=0x30 && wParam<=0x39) || (wParam>=0x41 && wParam<=0x5A) ) {
			(LRESULT)(!KEevent(KHDR_ALPHAKEY_UP, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		} else {
/****/		(LRESULT)(!KEevent(KHDR_KEY_UP, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		}
		return 0;
	 case WM_CHAR:
	// The KHDR_ALPHAKEY_DOWN and KHDR_ALPHAKEY_UP events are necessary in some
	// circumstances to make menus and dialogs work right with the keyboard, but
	// can be lived without if an OS doesn't make it easy to do them.  HOWEVER....
	// KHDR_CHAR *MUST* be sent during "normal typing" in order for "edit" fields
	// to work!!!
		if( wParam>=0x20 ) {
/****/		(LRESULT)(!KEevent(KHDR_CHAR, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0));
		}
		return 0;
	 case WM_MOUSEMOVE:
	// the mouse (pointer) has been moved
		CurCur = -1;	// preset cursor flag to "none set" so we know if KASM code sets it.
/****/	KEevent(KHDR_POINTER_MOVE, LOWORD(lParam), HIWORD(lParam), wParam);
		if( CurCur==-1 ) CurCur = 0;	// if KASM code didn't set cursor, force to "normal" cursor
		if( CurCur==2 ) SetCursor(LoadCursor(NULL, IDC_UPARROW));
		else if( CurCur==1 ) SetCursor(LoadCursor(NULL, IDC_WAIT));
		else SetCursor(LoadCursor(NULL, IDC_ARROW));
		return 0;
	 case WM_LBUTTONDOWN:
	// the LEFT mouse button has been pressed
/****/	KEevent(KHDR_KEY_DOWN, VK_LBUTTON, wParam, lParam);
		return 0;
	 case WM_LBUTTONUP:
	// the LEFT mouse button has been released
/****/	KEevent(KHDR_KEY_UP, VK_LBUTTON, wParam, lParam);
		return 0;
	 case WM_MBUTTONDOWN:
	// the MIDDLE mouse button has been pressed
/****/	KEevent(KHDR_KEY_DOWN, VK_MBUTTON, wParam, lParam);
		return 0;
	 case WM_MBUTTONUP:
	// the MIDDLE mouse button has been released
/****/	KEevent(KHDR_KEY_UP, VK_MBUTTON, wParam, lParam);
		return 0;
	 case WM_RBUTTONDOWN:
	// the RIGHT mouse button has been pressed
/****/	KEevent(KHDR_KEY_DOWN, VK_RBUTTON, wParam, lParam);
		return 0;
	 case WM_RBUTTONUP:
	// the RIGHT mouse button has been released
/****/	KEevent(KHDR_KEY_UP, VK_RBUTTON, wParam, lParam);
		return 0;
	 case WM_PAINT:
	// the program window needs updating, repaint it from the kBM bitmap
		hdc = BeginPaint(hWnd, &ps);

		SelectPalette(hdc, kPal, 0);
		RealizePalette(hdc);	// for 256 color displays

		SetBkColor(hdc, PALETTERGB(255,255,255));		
		SetTextColor(hdc, PALETTERGB(0,0,0));

		bdc=CreateCompatibleDC(hdc);
		holdbm = SelectObject(bdc, kBM->hBitmap);
		SelectPalette(bdc, kPal, 0);
		RealizePalette(bdc);

		BitBlt(hdc, 0, 0, VW, VH, bdc, 0, 0, SRCCOPY);

		SelectObject(bdc, oldBMs);
		DeleteDC(bdc);

		EndPaint(hWnd, &ps);
		return 0;
	 case WM_CLOSE:
	// someone has requested the closing of the program.
	// if KEevent(KHDR_QUERY_QUIT) returns 0, don't quit program
	//   else go ahead and quit
/****/	if( !KEevent(KHDR_QUERY_QUIT, 0, 0, 0) ) return 0;
		break;
	 case WM_DESTROY:
	// the program is being killed, tell the KASM code so it can clean up
/****/	KEevent(KHDR_SHUT_DOWN, 0, 0, 0);
	// kill the timer if the Kode failed to
		if( TimerID!=NULL ) KillTimer(hWndMain, TimerID);
		SysCleanup();

		DeleteDC(sdc);
		DeleteDC(ddc);

		PostQuitMessage(0);
		return 0;
	}
	return( DefWindowProc(hWnd, msg, wParam, lParam));
}

//**********************************************************************
//* Called from PMACH.CPP when the .KE wants the program to terminate
//* Do whatever's necessary to make the program terminate.
//*
void KillProgram() {

	DestroyWindow(hWndMain);
}

//***************
//* END OF FILE *
//***************
